/*
 * Designed and developed in 2022-2024 by William Walker (wnwalker@ucsc.edu)
 * Copyright 2022-2024 The Regents of the University of California. All rights reserved.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package edu.ucsc.its.temerity.shared.ui

import androidx.compose.animation.AnimatedVisibility
import androidx.compose.animation.fadeIn
import androidx.compose.animation.fadeOut
import androidx.compose.foundation.background
import androidx.compose.foundation.border
import androidx.compose.foundation.layout.Arrangement
import androidx.compose.foundation.layout.Box
import androidx.compose.foundation.layout.Column
import androidx.compose.foundation.layout.Row
import androidx.compose.foundation.layout.Spacer
import androidx.compose.foundation.layout.fillMaxHeight
import androidx.compose.foundation.layout.fillMaxSize
import androidx.compose.foundation.layout.fillMaxWidth
import androidx.compose.foundation.layout.height
import androidx.compose.foundation.layout.padding
import androidx.compose.foundation.layout.width
import androidx.compose.foundation.lazy.LazyColumn
import androidx.compose.foundation.lazy.itemsIndexed
import androidx.compose.foundation.shape.RoundedCornerShape
import androidx.compose.foundation.text.KeyboardActions
import androidx.compose.foundation.text.KeyboardOptions
import androidx.compose.material.icons.Icons.Outlined
import androidx.compose.material.icons.Icons.Rounded
import androidx.compose.material.icons.outlined.AdUnits
import androidx.compose.material.icons.outlined.Contacts
import androidx.compose.material.icons.outlined.Edit
import androidx.compose.material.icons.outlined.Group
import androidx.compose.material.icons.outlined.Notifications
import androidx.compose.material.icons.outlined.Settings
import androidx.compose.material.icons.rounded.Clear
import androidx.compose.material.icons.rounded.Search
import androidx.compose.material3.ExtendedFloatingActionButton
import androidx.compose.material3.FloatingActionButton
import androidx.compose.material3.HorizontalDivider
import androidx.compose.material3.Icon
import androidx.compose.material3.IconButton
import androidx.compose.material3.MaterialTheme
import androidx.compose.material3.NavigationBar
import androidx.compose.material3.NavigationBarItem
import androidx.compose.material3.NavigationDrawerItem
import androidx.compose.material3.OutlinedTextField
import androidx.compose.material3.OutlinedTextFieldDefaults
import androidx.compose.material3.PermanentDrawerSheet
import androidx.compose.material3.PermanentNavigationDrawer
import androidx.compose.material3.Scaffold
import androidx.compose.material3.Text
import androidx.compose.material3.minimumInteractiveComponentSize
import androidx.compose.material3.windowsizeclass.ExperimentalMaterial3WindowSizeClassApi
import androidx.compose.material3.windowsizeclass.WindowWidthSizeClass
import androidx.compose.material3.windowsizeclass.calculateWindowSizeClass
import androidx.compose.runtime.Composable
import androidx.compose.runtime.LaunchedEffect
import androidx.compose.runtime.getValue
import androidx.compose.runtime.mutableIntStateOf
import androidx.compose.runtime.mutableStateOf
import androidx.compose.runtime.remember
import androidx.compose.runtime.setValue
import androidx.compose.ui.Alignment
import androidx.compose.ui.Modifier
import androidx.compose.ui.draw.clip
import androidx.compose.ui.graphics.vector.ImageVector
import androidx.compose.ui.text.input.ImeAction
import androidx.compose.ui.text.input.KeyboardType
import androidx.compose.ui.unit.dp
import androidx.lifecycle.viewModelScope
import androidx.navigation.NavGraph
import androidx.navigation.NavHostController
import androidx.navigation.compose.composable
import androidx.navigation.compose.rememberNavController
import androidx.navigation.createGraph
import androidx.navigation.toRoute
import com.composables.composetheme.ComposeTheme
import com.composables.composetheme.material3.colorScheme
import edu.ucsc.its.temerity.Temerity
import edu.ucsc.its.temerity.model.Course
import edu.ucsc.its.temerity.model.Device
import edu.ucsc.its.temerity.model.User
import edu.ucsc.its.temerity.shared.ui.MainContentWindow.Courses
import edu.ucsc.its.temerity.shared.ui.MainContentWindow.Devices
import edu.ucsc.its.temerity.shared.ui.MainContentWindow.Notifications
import edu.ucsc.its.temerity.shared.ui.MainContentWindow.Roster
import edu.ucsc.its.temerity.shared.ui.MainContentWindow.Settings
import edu.ucsc.its.temerity.shared.ui.UserDetailView.BlankScreen
import edu.ucsc.its.temerity.shared.ui.UserDetailView.EditScreen
import edu.ucsc.its.temerity.shared.ui.elements.CoursesList
import edu.ucsc.its.temerity.shared.ui.elements.DeviceList
import edu.ucsc.its.temerity.shared.ui.elements.NewNavHost
import edu.ucsc.its.temerity.shared.ui.elements.UserList
import edu.ucsc.its.temerity.shared.viewmodel.AppViewModel
import kotlinx.coroutines.Dispatchers.IO
import kotlinx.coroutines.launch
import kotlinx.coroutines.withContext
import kotlinx.serialization.Serializable
import kotlinx.serialization.encodeToString
import kotlinx.serialization.json.Json
import org.dotenv.vault.dotenvVault
import org.jetbrains.compose.resources.StringResource
import org.jetbrains.compose.resources.stringResource
import org.koin.compose.getKoin
import org.koin.compose.viewmodel.koinViewModel
import org.koin.core.annotation.KoinExperimentalAPI
import org.koin.core.parameter.parametersOf
import temerity.shared.compose.generated.resources.Res.string
import temerity.shared.compose.generated.resources.courses
import temerity.shared.compose.generated.resources.devices
import temerity.shared.compose.generated.resources.notifications
import temerity.shared.compose.generated.resources.notifications_short
import temerity.shared.compose.generated.resources.roster
import temerity.shared.compose.generated.resources.settings

sealed class MainContentWindow {
  @Serializable
  data object Notifications : MainContentWindow()

  @Serializable
  data object Devices : MainContentWindow()

  @Serializable
  data object Roster : MainContentWindow()

  @Serializable
  data object Courses : MainContentWindow()

  @Serializable
  data object Settings : MainContentWindow()
}

sealed class UserDetailView {
  @Serializable
  data object BlankScreen : UserDetailView()

  @Serializable
  data class EditScreen(val user: String) : UserDetailView()
}

@OptIn(ExperimentalMaterial3WindowSizeClassApi::class, KoinExperimentalAPI::class)
@Composable
fun MainScreen(modifier: Modifier = Modifier) {
  val appViewModel: AppViewModel = koinViewModel()
  val appViewModelScope = appViewModel.viewModelScope
  // Set initial main content pane to "Devices"
  var navigationIndex by remember { mutableIntStateOf(1) }
  val windowSizeClass = calculateWindowSizeClass()
  val widthSizeClass = windowSizeClass.widthSizeClass

  val dotenv = dotenvVault {
    directory = "../temerity"
  }
  val pc: Temerity = getKoin().get(parameters = { parametersOf(dotenv["YUJAPROD_API_URL"], dotenv["YUJAPROD_TOKEN"]) })
  var devicesList by remember { mutableStateOf(listOf<Device>()) }
  var usersList by remember { mutableStateOf(listOf<User>()) }
  var coursesList by remember { mutableStateOf(listOf<Course>()) }
  var searchQuery by remember { mutableStateOf(String()) }

  LaunchedEffect(pc) {
    appViewModelScope.launch {
      withContext(IO) {
        devicesList = pc.getDevices().sortedBy { it.stationName }
        usersList = pc.getUsers().sortedBy { it.lastName }
        coursesList = pc.getCourses()
      }
    }
  }

  val userDetailViewController = rememberNavController()
  val userDetailViewGraph = userDetailViewController.createGraph(
    startDestination = BlankScreen,
  ) {
    composable<BlankScreen> {
      Box {
      }
    }
    composable<EditScreen> {
      val args = it.toRoute<EditScreen>()
      val user = Json.decodeFromString<User>(args.user)
      Box {
        UserEditScreen(
          user = user,
        )
      }
    }
  }

  val contentPaneNavController = rememberNavController()
  val contentPaneNavGraph = contentPaneNavController.createGraph(
    startDestination = Devices,
  ) {
    composable<Notifications> {
      // NotificationsScreen()
    }
    composable<Devices> {
      DeviceList(
        devicesList,
      )
    }
    composable<Roster> {
      UserList(
        usersList,
      ) {
        userDetailViewController.popBackStack()
        userDetailViewController.navigate(EditScreen(Json.encodeToString(it)))
      }
    }
    composable<Courses> {
      CoursesList(
        coursesList,
      ) {
      }
    }
    composable<Settings> {
      // SettingsScreen()
    }
  }

  val navigationItems = listOf(
    mapOf("icon" to Outlined.Notifications, "label" to if (widthSizeClass == WindowWidthSizeClass.Compact) string.notifications_short else string.notifications, "route" to Notifications),
    mapOf("icon" to Outlined.AdUnits, "label" to string.devices, "route" to Devices),
    mapOf("icon" to Outlined.Contacts, "label" to string.roster, "route" to Roster),
    mapOf("icon" to Outlined.Group, "label" to string.courses, "route" to Courses),
    mapOf("icon" to Outlined.Settings, "label" to string.settings, "route" to Settings),
  )

  if (widthSizeClass == WindowWidthSizeClass.Compact) {
    CompactScreen(
      modifier = modifier,
      navigationIndex = navigationIndex,
      navigationItems = navigationItems,
      contentPaneNavController = contentPaneNavController,
      contentPaneNavGraph = contentPaneNavGraph,
      onSelectNavigationMenuItem = { index, route ->
        navigationIndex = index
        contentPaneNavController.popBackStack()
        contentPaneNavController.navigate(route)
      },
    )
  } else {
    ExpandedScreen(
      modifier = modifier,
      navigationIndex = navigationIndex,
      navigationItems = navigationItems,
      contentPaneNavController = contentPaneNavController,
      contentPaneNavGraph = contentPaneNavGraph,
      userDetailViewController = userDetailViewController,
      userDetailViewGraph = userDetailViewGraph,
      widthSizeClass = widthSizeClass,
      onSelectNavigationMenuItem = { index, route ->
        navigationIndex = index
        contentPaneNavController.popBackStack()
        contentPaneNavController.navigate(route)
      },
    )
  }
}

@Composable
fun CompactScreen(
  navigationIndex: Int,
  navigationItems: List<Map<String, Any>>,
  contentPaneNavController: NavHostController,
  contentPaneNavGraph: NavGraph,
  modifier: Modifier = Modifier,
  onSelectNavigationMenuItem: (Int, MainContentWindow) -> Unit,
) {
  Scaffold(modifier = Modifier.fillMaxHeight(), floatingActionButton = {
    FloatingActionButton(onClick = { /* TODO */ }) {
      Icon(Outlined.Edit, "Compose new message")
    }
  }, bottomBar = {
    NavigationBar {
      navigationItems.forEachIndexed { index, _ ->
        val item = navigationItems[index]
        val icon = item["icon"] as ImageVector
        val label = item["label"] as StringResource
        val route = item["route"] as MainContentWindow
        NavigationBarItem(selected = index == navigationIndex, icon = { Icon(icon, null) }, label = { Text(stringResource(label)) }, onClick = {
          if (index != navigationIndex) {
            onSelectNavigationMenuItem(index, route)
          }
        })
      }
    }
  }) { padding ->
    ContentPanel(
      contentPaneNavController = contentPaneNavController,
      contentPaneNavGraph = contentPaneNavGraph,
      modifier = Modifier.fillMaxSize().padding(padding),
    )
  }
}

@Composable
fun ExpandedScreen(
  navigationIndex: Int,
  navigationItems: List<Map<String, Any>>,
  contentPaneNavController: NavHostController,
  contentPaneNavGraph: NavGraph,
  userDetailViewController: NavHostController,
  userDetailViewGraph: NavGraph,
  widthSizeClass: WindowWidthSizeClass,
  modifier: Modifier = Modifier,
  onSelectNavigationMenuItem: (Int, MainContentWindow) -> Unit,
) {
  Scaffold(modifier = modifier.fillMaxHeight()) {
    PermanentNavigationDrawer(drawerContent = {
      Row {
        PermanentDrawerSheet(Modifier.width(220.dp).padding(vertical = 24.dp)) {
          Column {
            ExtendedFloatingActionButton(
              onClick = { /*TODO*/ },
              modifier = Modifier.padding(horizontal = 12.dp),
            ) {
              Row(verticalAlignment = Alignment.CenterVertically, horizontalArrangement = Arrangement.spacedBy(8.dp, Alignment.Start)) {
                Icon(Outlined.Edit, null)
                Text("Compose")
              }
            }
            Spacer(Modifier.height(24.dp))
            LazyColumn {
              itemsIndexed(navigationItems) { index, _ ->
                val item = navigationItems[index]
                val icon = item["icon"] as ImageVector
                val label = item["label"] as StringResource
                val route = item["route"] as MainContentWindow
                NavigationDrawerItem(
                  modifier = Modifier.padding(horizontal = 16.dp),
                  selected = index == navigationIndex,
                  icon = { Icon(icon, null) },
                  label = { Text(stringResource(label)) },
                  onClick = {
                    if (index != navigationIndex) {
                      onSelectNavigationMenuItem(index, route)
                    }
                  },
                )
              }
            }
          }
        }
        HorizontalDivider(modifier = Modifier.width(1.dp).fillMaxHeight())
      }
    }, content = {
      Row {
        ContentPanel(
          contentPaneNavController = contentPaneNavController,
          contentPaneNavGraph = contentPaneNavGraph,
          modifier = Modifier.weight(1f).fillMaxHeight(),
        )
        if (widthSizeClass == WindowWidthSizeClass.Expanded) {
          HorizontalDivider(modifier = Modifier.width(1.dp).fillMaxHeight())
          DetailPanel(
            detailPaneViewController = userDetailViewController,
            detailPaneNavGraph = userDetailViewGraph,
            modifier = Modifier.weight(1f).padding(top = 76.dp),
          )
        }
      }
    })
  }
}

@Composable
fun SearchBar(
  searchQuery: String,
  onSearchQueryChange: (String) -> Unit,
  modifier: Modifier = Modifier,
  onImeSearch: () -> Unit,
) {
  Box(modifier = modifier.padding(horizontal = 16.dp, vertical = 10.dp)) {
    Row(horizontalArrangement = Arrangement.spacedBy(8.dp), verticalAlignment = Alignment.CenterVertically) {
      OutlinedTextField(
        colors = OutlinedTextFieldDefaults.colors(
          unfocusedContainerColor = ComposeTheme.colorScheme.surface,
          focusedContainerColor = ComposeTheme.colorScheme.surface,
        ),
        shape = RoundedCornerShape(100), placeholder = { Text("Search...") }, leadingIcon = {
          Icon(imageVector = Rounded.Search, contentDescription = null, modifier = Modifier.minimumInteractiveComponentSize(), tint = ComposeTheme.colorScheme.onSurface.copy(alpha = 0.66f))
        }, modifier = Modifier.weight(1f).minimumInteractiveComponentSize(), value = searchQuery, singleLine = true, keyboardActions = KeyboardActions(onSearch = { onImeSearch() }), keyboardOptions = KeyboardOptions(keyboardType = KeyboardType.Text, imeAction = ImeAction.Search), onValueChange = onSearchQueryChange, trailingIcon = {
          AnimatedVisibility(visible = searchQuery.isNotBlank(), enter = fadeIn(), exit = fadeOut()) {
            IconButton(onClick = { onSearchQueryChange("") }) {
              Icon(imageVector = Rounded.Clear, contentDescription = "Clear", tint = MaterialTheme.colorScheme.onSurface)
            }
          }
        },
      )
    }
  }
}

@Composable
fun ContentPanel(contentPaneNavController: NavHostController, contentPaneNavGraph: NavGraph, modifier: Modifier = Modifier) {
  var searchQuery = remember { mutableStateOf("") }
  Column(modifier) {
    SearchBar(
      searchQuery = searchQuery.value,
      onSearchQueryChange = {
        searchQuery.value = it
        contentPaneNavController.currentDestination!!.route?.let { currentNavDestination: String -> filterContent(currentNavDestination) }
      },
      onImeSearch = { /*TODO*/ },
    )
    Box(
      modifier = Modifier
        .weight(1f)
        .fillMaxWidth()
        .padding(16.dp)
        .border(1.dp, ComposeTheme.colorScheme.outline, RoundedCornerShape(18.dp))
        .clip(RoundedCornerShape(18.dp))
        .background(ComposeTheme.colorScheme.surface),
    ) {
      NewNavHost(
        navController = contentPaneNavController,
        navGraph = contentPaneNavGraph,
      )
    }
  }
}

@Composable
fun DetailPanel(detailPaneViewController: NavHostController, detailPaneNavGraph: NavGraph, modifier: Modifier = Modifier) {
  Box(
    modifier
      .fillMaxHeight()
      .padding(16.dp)
      .border(1.dp, ComposeTheme.colorScheme.outline, RoundedCornerShape(18.dp))
      .clip(RoundedCornerShape(18.dp))
      .background(ComposeTheme.colorScheme.surface),
  ) {
    NewNavHost(
      navController = detailPaneViewController,
      navGraph = detailPaneNavGraph,
    )
  }
}

fun filterContent(currentNavDestination: String) {
//  when(contentPaneNavController.findDestination(currentNavDestination)) {
//      is Devices -> {
//        devicesList = devicesList.filter { it.stationName.contains(searchQuery, ignoreCase = true) }
//      }
//      is Roster -> {
//        usersList = usersList.filter { it.loginId.contains(searchQuery, ignoreCase = true) }
//      }
//      is Courses -> {
//        coursesList = coursesList.filter { it.courseCode.contains(searchQuery, ignoreCase = true) }
//      }
//
//      Notifications -> TODO()
//      Settings -> TODO()
//  }
}
