/*
 *     Designed and developed in 2022-2024 by William Walker (wnwalker@ucsc.edu)
 *     Copyright 2022-2024 The Regents of the University of California. All rights reserved.
 *
 *     This library is free software; you can redistribute it and/or
 *     modify it under the terms of the GNU Lesser General Public
 *     License as published by the Free Software Foundation; version 2.1 of the License.
 *
 *     This library is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *     Lesser General Public License for more details.
 *
 *     You should have received a copy of the GNU Lesser General Public
 *     License along with this library; if not, write to the Free Software
 *     Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 */
package edu.ucsc.its.temerity.core

import co.touchlab.kermit.Logger
import com.skydoves.sandwich.ApiResponse
import com.skydoves.sandwich.StatusCode
import com.skydoves.sandwich.getOrThrow
import com.skydoves.sandwich.ktor.executeApiResponse
import com.skydoves.sandwich.ktor.statusCode
import com.skydoves.sandwich.onSuccess
import edu.ucsc.its.temerity.AuditLogSortOrder
import edu.ucsc.its.temerity.Temerity
import edu.ucsc.its.temerity.api.PlatformApi
import edu.ucsc.its.temerity.applyAuditLogFormat
import edu.ucsc.its.temerity.applyScheduledSessionDateFormat
import edu.ucsc.its.temerity.di.TemerityKoinComponent
import edu.ucsc.its.temerity.model.AuditLogEntry
import edu.ucsc.its.temerity.model.Course
import edu.ucsc.its.temerity.model.Device
import edu.ucsc.its.temerity.model.DeviceRecordingSession
import edu.ucsc.its.temerity.model.EventType
import edu.ucsc.its.temerity.model.FolderPermissions
import edu.ucsc.its.temerity.model.Group
import edu.ucsc.its.temerity.model.NewUser
import edu.ucsc.its.temerity.model.User
import edu.ucsc.its.temerity.model.UserGroup
import edu.ucsc.its.temerity.model.UserRecordingSession
import edu.ucsc.its.temerity.model.UserUpdate
import edu.ucsc.its.temerity.newJobScope
import edu.ucsc.its.temerity.sortByCreationDate
import io.ktor.client.statement.HttpResponse
import io.ktor.util.cio.toByteArray
import io.ktor.utils.io.ByteReadChannel
import kotlinx.coroutines.CompletableDeferred
import kotlinx.coroutines.async
import kotlinx.coroutines.awaitAll
import kotlinx.coroutines.delay
import kotlinx.coroutines.launch
import kotlinx.coroutines.runBlocking
import kotlinx.datetime.DateTimeUnit
import kotlinx.datetime.LocalDate
import kotlinx.datetime.minus
import kotlinx.serialization.SerializationException
import kotlinx.serialization.encodeToString
import kotlinx.serialization.json.Json
import org.koin.core.parameter.parametersOf

// Based on sample code available from https://github.com/santimattius/kmp-networking/blob/main/README.md
// Publicly instantiable only via factory method
public class TemerityImpl private constructor(
  private val debugEnabled: Boolean,
  private var platformApi: PlatformApi,
) : Temerity {

  /**
   * Exception class used to break out of a loop when reading paginated API responses.
   */
  private class BreakException(val page: Int) : RuntimeException() {
    fun lastPage(): Int = page
  }

  private val json = Json {
    coerceInputValues = true
  }

  /**
   * Builder class for Temerity
   */
  internal class Builder : TemerityKoinComponent(), Temerity.Builder {

    private var optDebugEnabled: Boolean = false

    /**
     * Specifies whether debug logging should be enabled
     * @param debugEnabled
     * Enable this for debugging purposes only
     */
    override fun debugEnabled(debugEnabled: Boolean) {
      optDebugEnabled = debugEnabled
    }

    private var serviceEndpoint: String = ""

    /**
     * Specifies URL of platform instance to make API calls to
     * @param endpoint The URL of the platform instance to make API calls to
     */
    override fun serviceEndpoint(endpoint: String) {
      serviceEndpoint = endpoint
    }

    private var serviceToken: String = ""

    /**
     * Specifies the authorization token to use for API calls
     *  @param token The authorization token to use for API calls
     */
    override fun serviceToken(token: String) {
      serviceToken = token
    }

    /**
     * Apply changes to builder and get the PlatformClient instance without needing to call [build] afterward.
     */

    @Suppress("CanBePrivate")
    override fun build(builder: Temerity.Builder.() -> Unit): Temerity = this.apply(builder).build()

    /**
     * Creates a Temerity instance with specified service endpoint and token.
     */
    override fun build(): Temerity {
      require(serviceEndpoint.isNotBlank()) {
        "Must pass platform service endpoint as parameter"
      }
      require(serviceToken.isNotBlank()) {
        "Must pass platform authorization token as parameter"
      }
      val koin = getKoin()
      val platformApi: PlatformApi = koin.get { parametersOf(serviceEndpoint, optDebugEnabled, serviceToken) }
      return TemerityImpl(
        debugEnabled = optDebugEnabled,
        platformApi,
      )
    }
  }

  // Inlining advantageous here since copied parameter (ApiResponse<String>) is small and code fragment is used constantly (to validate every platform API request)
  private inline fun <reified T> decodeResponseCatching(response: ApiResponse<String>, page: Int = 0): T {
    // Execute platform request, throwing any network exceptions
    // TODO: Handle some network exceptions here, such as timeouts (for retry) or unauthorized (for invalid token)
    response.onSuccess {
      when (this.statusCode) {
        // If the API returns a 204 No Content status code, we should throw a BreakException to stop the page iteration loop
        // As of Cashew release, this is only used in getGroups() to stop the loop when no more groups are returned
        // May be used in the future for other paginated endpoints (such as getAuditLogEntries)
        StatusCode.NoContent -> {
          throw BreakException(page - 1)
        }

        else -> Unit
      }
    }
    val apiResponsePayload = response.getOrThrow()
    // Handle deserialization exceptions. Queue up decode:
    val deserializedJson = runCatching {
      json.decodeFromString<T>(apiResponsePayload)
    }
    // Attempt, returning the deserialized object if successful
    return deserializedJson.getOrElse { exception: Throwable ->
      when (exception) {
        is SerializationException -> {
          // TODO: Implement platform API version checking
          error("Encountered error decoding response from platform API. You likely need to choose a Temerity release that supports the API version implemented by your instance. \nPlatform API response data : $apiResponsePayload")
        }

        else -> throw exception
      }
    }
  }

  private fun <T> getPaginatedItems(
    fetchPage: suspend (Int) -> ApiResponse<String>,
    delayMillis: Long = 750L,
    pageSize: Int = 75,
  ): List<T> {
    var returnedItems: List<T> = emptyList()
    val jobScope = newJobScope()
    val requestJobs = ArrayList<CompletableDeferred<List<T>>>()
    runBlocking {
      try {
        for (x in (0..Int.MAX_VALUE)) {
          val thisDeferredJob = CompletableDeferred<List<T>>()
          requestJobs.add(x, thisDeferredJob)
          jobScope.launch {
            val response = decodeResponseCatching<List<T>>(fetchPage(x*pageSize))
            thisDeferredJob.complete(response)
          }
          delay(delayMillis)
        }
      } catch (e: Exception) {
        when (e) {
          is BreakException -> {
            if (debugEnabled) {
              TemerityLibrary.koin.get<Logger>().d("Caught BreakException() from decodeResponse() notifying we're done reading items from API: $e")
            }
            returnedItems = requestJobs.awaitAll().flatten()
            return@runBlocking returnedItems
          }
          else -> throw e
        }
      }
      returnedItems = requestJobs.awaitAll().flatten()
    }
    return returnedItems
  }

  public override suspend fun getUsers(): List<User> =
    decodeResponseCatching(platformApi.getUsers().executeApiResponse<String>())

  public override suspend fun getUser(userId: Long): User =
    getUsers().first { it.userId == userId }

  public override suspend fun createUser(primaryIdentifier: String, newUser: NewUser): HttpResponse {
    val validationMessage = newUser.validate()
    if (validationMessage != "Valid") {
      error(validationMessage)
    } else {
      return platformApi.createUser(primaryIdentifier, json.encodeToString(newUser)).getOrThrow()
    }
  }

  public override suspend fun updateUser(userId: Long, userUpdate: UserUpdate): HttpResponse =
    platformApi.setUser(userId, json.encodeToString(userUpdate)).getOrThrow()

  public override suspend fun deleteUser(userId: Long): HttpResponse =
    platformApi.deleteUser(userId).getOrThrow()

  public override suspend fun getUserGroups(userId: Long): List<UserGroup> =
    decodeResponseCatching(platformApi.getUserGroups(userId).executeApiResponse<String>())

  public override suspend fun getUserGroupsOwned(userId: Long): List<UserGroup> =
    decodeResponseCatching<List<UserGroup>>(platformApi.getUserGroupsOwned(userId).executeApiResponse())

  public override suspend fun createUserWithExternalId(primaryIdentifier: String, newUser: NewUser): HttpResponse {
    TODO("Not yet implemented")
  }

  public override suspend fun addFolderPermission(perms: FolderPermissions.Action): HttpResponse {
    TODO("Not yet implemented")
  }

  public override suspend fun removeFolderPermission(perms: FolderPermissions.Action) {
    TODO("Not yet implemented")
  }

  public override suspend fun getMediaMetadata(linkageValue: String) {
    TODO("Not yet implemented")
  }

  public override suspend fun getGroups(): List<Group> {
    return getPaginatedItems(
      fetchPage = { page -> platformApi.getGroupPage(page, true) },
    )
  }

  public override suspend fun getCourse(courseCode: String): Course {
    return decodeResponseCatching(platformApi.getCourse(courseCode).executeApiResponse<String>())
  }

  public override suspend fun getCourses(paginated: Boolean): List<Course> {
    return when (paginated) {
      true -> {
        return getPaginatedItems(
          fetchPage = { page -> platformApi.getGroupPage(page, true) },
        )
      }

      false -> {
        decodeResponseCatching(platformApi.getCourses().executeApiResponse<String>())
      }
    }
  }

  // Get all audit log entries for a list of event types within window
  // Defaults to getting audit log entries for all event types
  // TODO: Read sort order from settings
  public override suspend fun getAuditLogEntries(
    startTime: LocalDate,
    endTime: LocalDate,
    eventTypeList: List<EventType>?,
    sortOrder: AuditLogSortOrder?,
  ): List<AuditLogEntry> {
    // TODO: Error out if user inputs time window > 1 month, since the platform API doesn't support
    val returnedEntries = ArrayList<AuditLogEntry>()
    val jobScope = newJobScope()
    val requestJobs = with(eventTypeList ?: EventType.entries) {
      map { eventType ->
        jobScope.async {
          getAuditLogEntries(startTime, endTime, eventType, sortOrder)
        }
      }
    }

    val results = requestJobs.awaitAll()
    results.forEach { returnedEntries.addAll(it) }

    // Apply passed sortOrder, or default to NEW_FIRST TODO: Read from user settings
    returnedEntries.applyOrDefault(sortOrder ?: AuditLogSortOrder.NEW_FIRST)
    return returnedEntries
  }

  // Get audit log entries for a specific event type within window
  // TODO: Read sort order from settings
  public override suspend fun getAuditLogEntries(
    startTime: LocalDate,
    endTime: LocalDate,
    eventType: EventType,
    sortOrder: AuditLogSortOrder?,
  ): List<AuditLogEntry> {
    val returnedEntries = ArrayList<AuditLogEntry>()
    var windowStart = endTime.minus(1, DateTimeUnit.Companion.MONTH)
    var windowEnd = endTime
    while (windowStart < windowEnd) {
      val endpointRequest = platformApi.getAuditLogEntries(
        startTime = windowStart.applyAuditLogFormat(),
        endTime = windowEnd.applyAuditLogFormat(),
        eventType = eventType.value,
      )
      val logEventEntries: List<AuditLogEntry> = decodeResponseCatching(endpointRequest.executeApiResponse<String>())
      returnedEntries.addAll(logEventEntries)
      windowEnd = windowStart
      val backStep = windowStart.minus(1, DateTimeUnit.Companion.MONTH)
      windowStart = if (backStep < startTime) startTime else backStep
    }
    // Apply passed sortOrder, or default to NEW_FIRST TODO: Read from user settings
    returnedEntries.applyOrDefault(sortOrder ?: AuditLogSortOrder.NEW_FIRST)
    return returnedEntries
  }

  private fun ArrayList<AuditLogEntry>.applyOrDefault(sortOrder: AuditLogSortOrder?): ArrayList<AuditLogEntry> {
    return when (sortOrder) {
      null -> {
        // Currently reads from prefs, falling back to BuildConfig constant. Can be overridden by passing a sort order as a parameter.
        // TODO: Implement a way to pass a clientDefaultSortOrder to Builder
        // TODO: Persist user-defined default sort order for subsequent runs
        apply { sortByCreationDate(AuditLogSortOrder.NEW_FIRST) }
      }

      else -> apply { sortByCreationDate(sortOrder) }
    }
  }

  // TODO: Implement setUser() function

  public override suspend fun getDevices(): List<Device> =
    decodeResponseCatching(platformApi.getDevices().executeApiResponse<String>())

  public override suspend fun getDevice(deviceId: Long): Device =
    decodeResponseCatching(platformApi.getDeviceById(deviceId).executeApiResponse<String>())

  public override suspend fun getUserSessions(
    userId: String,
    startTime: LocalDate,
    endTime: LocalDate,
  ): List<UserRecordingSession> =
    decodeResponseCatching(
      platformApi.getUserSessions(
        userId,
        startTime.applyScheduledSessionDateFormat(),
        endTime.applyScheduledSessionDateFormat(),
      ).executeApiResponse<String>(),
    )

  public override suspend fun getDeviceSchedule(
    deviceId: Long,
    startTime: LocalDate,
    endTime: LocalDate,
  ): List<DeviceRecordingSession> =
    decodeResponseCatching(
      platformApi.getDeviceSchedule(
        deviceId,
        startTime.applyScheduledSessionDateFormat(),
        endTime.applyScheduledSessionDateFormat(),
      ).executeApiResponse<String>(),
    )

  public override suspend fun getStorageAnalyticsReport(groupId: Long): ByteArray =
    platformApi.getStorageAnalyticsReport(groupId).executeApiResponse<ByteReadChannel>().getOrThrow().toByteArray(limit = 2000000000) // Limit file downloads to 2 GB
}
