/*
 *     Designed and developed in 2022-2024 by William Walker (wnwalker@ucsc.edu)
 *     Copyright 2022-2024 The Regents of the University of California. All rights reserved.
 *
 *     This library is free software; you can redistribute it and/or
 *     modify it under the terms of the GNU Lesser General Public
 *     License as published by the Free Software Foundation; version 2.1 of the License.
 *
 *     This library is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *     Lesser General Public License for more details.
 *
 *     You should have received a copy of the GNU Lesser General Public
 *     License along with this library; if not, write to the Free Software
 *     Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 */
package edu.ucsc.its.temerity.test
import edu.ucsc.its.temerity.AuditLogSortOrder.NEW_FIRST
import edu.ucsc.its.temerity.Temerity
import edu.ucsc.its.temerity.core.temerityInstance
import edu.ucsc.its.temerity.currentDate
import edu.ucsc.its.temerity.di.loggerModule
import edu.ucsc.its.temerity.model.EventType.AUTOMATED_SESSION_FAILED_TO_START
import edu.ucsc.its.temerity.model.EventType.AUTOMATED_SESSION_MONITOR
import edu.ucsc.its.temerity.model.EventType.CAPTURE_ERROR
import edu.ucsc.its.temerity.model.EventType.DEVICE_ERROR
import edu.ucsc.its.temerity.model.EventType.DEVICE_FAILED_TO_START
import edu.ucsc.its.temerity.model.EventType.ERROR
import edu.ucsc.its.temerity.model.EventType.HARDWARE_CAPTURE_HUB
import edu.ucsc.its.temerity.model.EventType.HDCP_SIGNAL_DETECTED
import edu.ucsc.its.temerity.model.EventType.HUB_CAPTURE_PROBLEM
import edu.ucsc.its.temerity.model.EventType.NEW_LOG_IN
import edu.ucsc.its.temerity.model.EventType.RECORDING_ERROR
import io.kotest.core.extensions.Extension
import io.kotest.core.spec.style.FunSpec
import io.kotest.matchers.file.shouldNotBeEmpty
import kotlinx.coroutines.runBlocking
import kotlinx.datetime.Clock
import kotlinx.datetime.DateTimeUnit
import kotlinx.datetime.TimeZone
import kotlinx.datetime.minus
import kotlinx.datetime.todayIn
import org.dotenv.vault.dotenvVault
import org.jetbrains.kotlinx.dataframe.api.toDataFrame
import org.jetbrains.kotlinx.dataframe.io.writeCSV
import org.koin.core.parameter.parametersOf
import org.koin.test.KoinTest
import org.koin.test.inject
import java.io.File
import kotlin.time.Duration.Companion.minutes
import kotlin.uuid.ExperimentalUuidApi
import kotlin.uuid.Uuid
import co.touchlab.kermit.Logger as KermitLogger

@OptIn(ExperimentalUuidApi::class)
class ProdReportTests : FunSpec(), KoinTest {

  override fun extensions(): List<Extension> = listOf(koinExtension(loggerModule()))
  private val kermit: KermitLogger by inject { parametersOf("ProdReportTests") }

  init {
    coroutineDebugProbes = true

    val dotenv = dotenvVault()

    lateinit var testTemerity: Temerity
    val today = Clock.System.todayIn(TimeZone.currentSystemDefault())

    beforeTest {
      testTemerity = temerityInstance {
        debugEnabled(true)
        serviceEndpoint(dotenv["YUJAPROD_API_URL"])
        serviceToken(dotenv["YUJAPROD_TOKEN"])
      }
    }

    timeout = 5.minutes.inWholeMilliseconds

    test("PlatformClient can fetch a list of devices") {
      runBlocking {
        val returnedDevices = testTemerity.getDevices()
        println("Returned devices:")
        returnedDevices.forEach { println(it) }
        val df = returnedDevices.toDataFrame()
        df.writeCSV("YuJaDevices-${currentDate()}.csv")
        assert(returnedDevices.isNotEmpty())
      }
    }

    // Generate Report listing all groups
    test("PlatformClient can generate Groups report") {
      runBlocking {
        val returnedGroups = testTemerity.getGroups()
        println("Total returned groups count: ${returnedGroups.size}")
        val df = returnedGroups.toDataFrame()
        df.writeCSV("YuJaGroups.csv")
        assert(returnedGroups.isNotEmpty())
      }
    }

    test("PlatformClient can generate Group terms list report") {
      runBlocking {
        val returnedGroups = testTemerity.getGroups()
        println("Total returned groups count: ${returnedGroups.size}")
        val termList = returnedGroups.map { it.groupTerm }.distinct()
        val df = termList.toDataFrame()
        df.writeCSV("YuJaTermsList.csv")
        assert(returnedGroups.isNotEmpty())
      }
    }

    test("PlatformClient can generate User list report") {
      runBlocking {
        val returnedUsers = testTemerity.getUsers()
        println("Total returned user count: ${returnedUsers.size}")
        val df = returnedUsers.toDataFrame()
        df.writeCSV("YuJaUsersList-${currentDate()}-${Uuid.random()}.csv")
        assert(returnedUsers.isNotEmpty())
      }
    }

    // Generate Report listing all groups matching specific term code
    test("PlatformClient can fetch a list of groups from Spring Quarter 2024") {
      runBlocking {
        val returnedGroups = testTemerity.getGroups().filter { it.groupTerm == "2024 Spring Quarter" }
        println("Total returned groups count: ${returnedGroups.size}")
        returnedGroups.forEach { println(it) }
        val df = returnedGroups.toDataFrame()
        df.writeCSV("2024SpringQuarterGroups.csv")
        assert(returnedGroups.isNotEmpty())
      }
    }

    // Generate Report listing all groups matching specific term code
    test("PlatformClient can fetch a list of courses from Fall Quarter 2024") {
      runBlocking {
        val returnedCourses = testTemerity.getCourses().filter { it.courseTerm == "2024 Fall Quarter" }
        println("Total returned courses count: ${returnedCourses.size}")
        returnedCourses.forEach { println(it) }
        val df = returnedCourses.toDataFrame()
        df.writeCSV("2024FallQuarterCourses.csv")
        assert(returnedCourses.isNotEmpty())
      }
    }

    test("PlatformClient can fetch a list of groups from sis term code 2248") {
      runBlocking {
        val returnedCourses = testTemerity.getCourses().filter { it.sisId.startsWith("2248") }
        println("Total returned groups count: ${returnedCourses.size}")
        returnedCourses.forEach { println(it) }
        val df = returnedCourses.toDataFrame()
        df.writeCSV("2024FallQuarterGroups-${currentDate()}.csv")
        assert(returnedCourses.isNotEmpty())
      }
    }

    test("PlatformClient can fetch a list of groups from sis term code 2240 and verify term code matches") {
      runBlocking {
        val returnedGroups = testTemerity.getGroups().filter {
          it.sisId.startsWith("2240")
        }.filter {
          it.groupTerm != "2024 Winter Quarter"
        }
        println("Total returned groups count: ${returnedGroups.size}")
        returnedGroups.forEach { println(it) }
        val df = returnedGroups.toDataFrame()
        df.writeCSV("2024WinterQuarterGroups-WrongTerm-${currentDate()}.csv")
        assert(returnedGroups.isNotEmpty())
      }
    }

    // "Mislinked" YuJa course group report
    // Generate Report listing all groups matching specific sis id that have the wrong term code, or are duplicates
    test("PlatformClient can fetch a list of groups with specific SIS ID by term and verifying term code matches") {
      runBlocking {
        val fullCoursesList = testTemerity.getCourses()
        val invalidGroupTerm = fullCoursesList.filter {
          it.sisId.startsWith("2248")
        }.filter {
          it.courseTerm != "2024 Fall Quarter"
        }
        val nonMatchingSisId = fullCoursesList.filter {
          it.courseTerm == "2024 Fall Quarter"
        }.filter {
          !it.sisId.startsWith("2248")
        }
        val duplicateSisId = fullCoursesList.filter {
          it.sisId.startsWith("2248")
        }.groupBy { it.sisId }.filter { it.value.size > 1 }.values.flatten()
        val problemCourses = invalidGroupTerm + nonMatchingSisId + duplicateSisId
        println("Total returned groups count: ${problemCourses.size}")
        problemCourses.forEach { kermit.i(it.toString()) }
        val df = problemCourses.toDataFrame()
        df.writeCSV("wrong-term-duplicate-course-report-${Uuid.random()}-${currentDate()}.csv")
      }
    }

    test("YuJaClient can fetch the whole list of groups") {
      runBlocking {
        val returnedGroups = testTemerity.getGroups()
        kermit.i("Total returned groups count: ${returnedGroups.size}")
        returnedGroups.forEach { println(it) }
        assert(returnedGroups.isNotEmpty())
      }
    }

    test("Get all audit log entries of type \"New Login\" from the past day") {
      runBlocking {
        val testEventType = NEW_LOG_IN
        val currentDate = Clock.System.todayIn(TimeZone.currentSystemDefault())
        val pastDate = currentDate.minus(1, DateTimeUnit.DAY)

        val returnedAuditLogEntries = testTemerity.getAuditLogEntries(
          startTime = pastDate,
          endTime = currentDate,
          eventType = testEventType,
          sortOrder = NEW_FIRST,
        )
        returnedAuditLogEntries.forEach { println(it) }
        val df = returnedAuditLogEntries.toDataFrame()

        df.writeCSV("${testEventType.name}-Event-Report-${Uuid.random()}-$pastDate-$currentDate-generated-${currentDate()}.csv")
        assert(returnedAuditLogEntries.isNotEmpty())
      }
    }

    test("Get and compile device audit log entries of multiple types from the past 2 years") {
      runBlocking {
        val testEventTypeList = listOf(
          AUTOMATED_SESSION_FAILED_TO_START,
          ERROR,
          RECORDING_ERROR,
          DEVICE_FAILED_TO_START,
          DEVICE_ERROR,
          HUB_CAPTURE_PROBLEM,
          HDCP_SIGNAL_DETECTED,
          CAPTURE_ERROR,
          HARDWARE_CAPTURE_HUB,
        )
        val pastDate = today.minus(2, DateTimeUnit.YEAR)
        val logEntries = testTemerity.getAuditLogEntries(
          startTime = pastDate,
          endTime = today,
          eventTypeList = testEventTypeList,
        )
        val df = logEntries.toDataFrame()
        val saveFilename = "DeviceAndSessionError-Event-Report-$pastDate-$today-generated-${currentDate()}-${Uuid.random()}.csv"
        df.writeCSV(saveFilename)
        File(saveFilename).shouldNotBeEmpty()
      }
    }

    test("Dump all audit log reports of type \"Automated Session Monitor\" from the past week") {
      runBlocking {
        val testEventType = AUTOMATED_SESSION_MONITOR
        val windowStart = today.minus(1, DateTimeUnit.WEEK)
        val logEntries = testTemerity.getAuditLogEntries(
          startTime = windowStart,
          endTime = today,
          eventType = testEventType,
        )
        val df = logEntries.toDataFrame()
        val saveFilename = "${testEventType.name}-Event-Report-${Uuid.random()}-$windowStart-$today-generated-${currentDate()}.csv"
        df.writeCSV(saveFilename)
        File(saveFilename).shouldNotBeEmpty()
      }
    }
  }
}
